#include <lektor/queue.h>
#include <lektor/common.h>
#include <lektor/database.h>
#include <lektor/mkv.h>
#include <lektor/internal/dbmacro.h>
#include <lektor/stb/ds.h>

#define sqlite_just_exec(func, query)                                 \
    bool func(lkt_db *db)                                             \
    {                                                                 \
        SQLITE_EXEC(db, "BEGIN TRANSACTION;" query "COMMIT;", error); \
        return true;                                                  \
    error:                                                            \
        SQLITE_DO_ROLLBACK(db);                                       \
        return false;                                                 \
    }
// clang-format off
sqlite_just_exec(database_queue_toggle_pause, "UPDATE queue_state SET paused = 1 - paused;");
sqlite_just_exec(database_queue_crop, "DELETE FROM queue WHERE queue.kara_id <> (SELECT current FROM queue_state LIMIT 1);");
sqlite_just_exec(database_queue_stop, "UPDATE queue_state SET current = NULL;");
sqlite_just_exec(database_queue_clear, "DELETE FROM queue;"
                                       "DELETE FROM queue_tmp;"
                                       "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
                                       "DELETE FROM " LKT_PROTECTED_DATABASE ".sqlite_sequence WHERE name = 'queue';"
                                       "UPDATE queue_state SET current = NULL;");
sqlite_just_exec(database_config_queue_default, "UPDATE queue_state SET volume = 100, paused = 1,"
                                                " random = 0, repeat = 0, single = 0, consume = 0,"
                                                " current = NULL, duration = 0;");
sqlite_just_exec(database_queue_flat, "UPDATE queue SET priority = 1;");
// clang-format on
#undef sqlite_just_exec

PRIVATE_FUNCTION void
queue_resequence__(lkt_db *db)
{
    static const char *SQL =
        "UPDATE " LKT_PROTECTED_DATABASE
        ".sqlite_sequence SET seq = (SELECT COUNT(*) FROM queue) WHERE name = 'queue';";
    SQLITE_EXEC(db, SQL, error);
    LOG_INFO("DB", "Re-sequenced the queue");
    return;
error:
    LOG_ERROR("DB", "Failed to update the " LKT_PROTECTED_DATABASE ".sqlite_sequence table...");
    return;
}

bool
database_queue_set_paused(lkt_db *db, bool paused)
{
    const char *SQL =
        paused ? "UPDATE queue_state SET paused = 1;" : "UPDATE queue_state SET paused = 0;";

    SQLITE_EXEC(db, SQL, error);
    return true;
error:
    return false;
}

bool
database_queue_state(lkt_db *db, struct lkt_queue_state *res)
{
    static const char *SQL_STMT =
        "SELECT"
        " volume, paused, random, repeat, single, current, duration, consume,"
        " (SELECT COUNT(*) FROM queue) AS length "
        "FROM queue_state;\n";
    sqlite3_stmt *stmt = 0;
    bool ret           = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_STEP_ROW(db, stmt, error);

    res->volume = sqlite3_column_int(stmt, 0);
    res->paused = sqlite3_column_int(stmt, 1);
    res->random = sqlite3_column_int(stmt, 2);
    res->repeat = sqlite3_column_int(stmt, 3);
    res->single = sqlite3_column_int(stmt, 4);

    if (sqlite3_column_type(stmt, 5) == SQLITE_NULL)
        res->current = -1;
    else
        res->current = sqlite3_column_int(stmt, 5);

    if (sqlite3_column_type(stmt, 6) == SQLITE_NULL)
        res->duration = 0;
    else
        res->duration = sqlite3_column_int(stmt, 6);

    res->consume = sqlite3_column_int(stmt, 7);
    res->length  = sqlite3_column_int(stmt, 8);
    ret          = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

void
database_queue_playtime(lkt_db *db, uint64_t *seconds)
{
    static const char *SQL = "SELECT SUM(cached_duration) "
                             "FROM kara "
                             "JOIN queue ON queue.kara_id = kara.id;";
    sqlite3_stmt *stmt     = NULL;
    *seconds               = 0;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *seconds = (uint64_t)sqlite3_column_int(stmt, 0);
error:
    sqlite3_finalize(stmt);
    return;
}

void
database_total_playtime(lkt_db *db, uint64_t *seconds)
{
    static const char *SQL = "SELECT SUM(cached_duration) FROM kara;";
    sqlite3_stmt *stmt     = NULL;
    *seconds               = 0;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *seconds = (uint64_t)sqlite3_column_int(stmt, 0);
error:
    sqlite3_finalize(stmt);
    return;
}

bool
database_queue_current_kara(lkt_db *db, struct kara_metadata *res, int *id)
{
    static const char *SQL_STMT = "SELECT song_name, source_name, category, language, author_name, "
                                  "song_type, song_number, kara_id"
                                  " FROM kara"
                                  " JOIN queue ON kara_id = kara.id"
                                  " JOIN queue_state ON current = position";
    sqlite3_stmt *stmt          = NULL;
    int ret                     = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_STEP_ROW(db, stmt, error);

    /* Here use gotos because of optimisations done by compilators.
       Most of the time it won't be NULL. */
    if (!res)
        goto no_metadata;
    // clang-format off
    strncpy(res->song_name,   sqlite3_column_chars(stmt, 0), LEKTOR_TAG_MAX - 1);
    strncpy(res->source_name, sqlite3_column_chars(stmt, 1), LEKTOR_TAG_MAX - 1);
    strncpy(res->category,    sqlite3_column_chars(stmt, 2), LEKTOR_TAG_MAX - 1);
    strncpy(res->language,    sqlite3_column_chars(stmt, 3), LEKTOR_TAG_MAX - 1);
    strncpy(res->author_name, sqlite3_column_chars(stmt, 4), LEKTOR_TAG_MAX - 1);
    strncpy(res->song_type,   sqlite3_column_chars(stmt, 5), LEKTOR_TAG_MAX - 1);
    // clang-format on
    res->song_number = sqlite3_column_int(stmt, 6);
no_metadata:
    /* Most of the time this will be NULL. */
    if (id && sqlite3_column_type(stmt, 7) != SQLITE_NULL)
        *id = sqlite3_column_int(stmt, 7);

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

#define reorder(db, prio, error)                        \
    if (prio > 1) {                                     \
        if (__queue_reorder(db)) {                      \
            LOG_INFO("DB", "Queue has been reordered"); \
        } else {                                        \
            LOG_INFO("DB", "Failed to reorder");        \
            goto error;                                 \
        }                                               \
    }

static bool
__queue_reorder(lkt_db *db)
{
#define CURRENT_POS_OR_0 /* To move karas after the current */           \
    "(SELECT CASE WHEN (SELECT current FROM queue_state) IS NULL THEN 0" \
    " ELSE (SELECT current FROM queue_state) END AS val LIMIT 1)"
    static const char *SQL_REORDER =
        /* Clear the TMP */
        "DELETE FROM queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
        /* Separate karas that are after the current one */
        "INSERT INTO queue_tmp (kara_id, priority)"
        " SELECT kara_id, priority FROM queue WHERE position > " CURRENT_POS_OR_0
        " ORDER BY priority DESC, position ASC;"
        "DELETE FROM queue WHERE position > " CURRENT_POS_OR_0 ";"
        /* Update the sqlite_sequence table */
        "UPDATE " LKT_PROTECTED_DATABASE ".sqlite_sequence SET seq = " CURRENT_POS_OR_0
        " WHERE name = 'queue';"
        /* Insert back */
        "INSERT INTO queue (position, kara_id, priority)"
        " SELECT position + " CURRENT_POS_OR_0 ", kara_id, priority"
        " FROM queue_tmp;"
        /* Clear the TMP */
        "DELETE FROM queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';";
#undef CURRENT_POS_OR_0
    SQLITE_EXEC(db, SQL_REORDER, error);
    return true;
error:
    return false;
}

static bool
queue_add_with_col_like_str(lkt_db *db, const char *col, const char *val, int priority)
{
    static const char *SQL_STMT =
        /* Add correspondig karas into the queue */
        "INSERT INTO queue (kara_id, priority)"
        " SELECT id, ?"
        " FROM kara"
        " WHERE %s LIKE ? AND available = 1"
        " ORDER BY RANDOM();";

    RETURN_IF(___is_sql_str_invalid(col), "Column name is invalid", false);
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt = NULL;

    /* Begin, transactions are atomic so we won't be anoyed
       by multi-threading */
    SQLITE_EXEC(db, "BEGIN TRANSACTION;", error);

    /* Insert at the end of the queue */
    safe_snprintf(SQL, LKT_MAX_SQLITE_STATEMENT, SQL_STMT, col);
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, priority, error);
    SQLITE_BIND_TEXT(db, stmt, 2, val, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sqlite3_finalize(stmt);

    /* Reorder kara that are after the current one, do this shit only
       if priority > 1... */
    reorder(db, priority, error);

    /* End */
    SQLITE_EXEC(db, "COMMIT;", error);
    LOG_INFO("DB-DEBUG", "Added with col %s like %s with priority %d", col, val, priority);
    return true;

error:
    sqlite3_finalize(stmt);
    SQLITE_DO_ROLLBACK(db);
    queue_resequence__(db);
    return false;
}

PRIVATE_FUNCTION bool
database_queue_add_plt(lkt_db *db, const char *plt_name, int priority)
{
    static const char *SQL_STMT =
        /* Insert the playlist */
        "INSERT INTO queue (kara_id, priority) "
        "SELECT kara.id, ?"
        " FROM kara"
        " JOIN kara_playlist ON kara_id = kara.id AND kara.available = 1"
        " JOIN playlist ON playlist_id = playlist.id AND playlist.name = ?"
        " ORDER BY RANDOM();";
    sqlite3_stmt *stmt = NULL;

    /* Begin, transactions are atomic so we won't be anoyed
       by multi-threading */
    SQLITE_EXEC(db, "BEGIN TRANSACTION;", error);

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, priority, error);
    SQLITE_BIND_TEXT(db, stmt, 2, plt_name, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sqlite3_finalize(stmt);

    /* Do the move shit only if the priority > 1 */
    reorder(db, priority, error);

    SQLITE_EXEC(db, "COMMIT;", error);
    LOG_INFO("DB-DEBUG", "Added play list %s with priority %d", plt_name, priority);
    return true;

error:
    sqlite3_finalize(stmt);
    SQLITE_DO_ROLLBACK(db);
    queue_resequence__(db);
    return false;
}

bool
database_queue_add_id(lkt_db *db, int id, int priority)
{
    static const char *SQL_STMT = "INSERT INTO queue (kara_id, priority)"
                                  " SELECT kara.id, ?"
                                  " FROM kara"
                                  " WHERE kara.id = ?;";
    sqlite3_stmt *stmt          = NULL;
    SQLITE_EXEC(db, "BEGIN TRANSACTION;", error);
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, priority, error);
    SQLITE_BIND_INT(db, stmt, 2, id, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sqlite3_finalize(stmt);

    /* Do the move shit only if the priority > 1 */
    reorder(db, priority, error_no_stmt);

    SQLITE_EXEC(db, "COMMIT;", error);
    return true;
error:
    LOG_ERROR("DB",
              "Error while adding the kara id %d into the queue with the "
              "priority %d",
              id, priority);
    sqlite3_finalize(stmt);
error_no_stmt:
    LOG_ERROR("DB",
              "Need to rollback the queue because of failed insertion of id "
              "%d with priority %d in the queue",
              id, priority);
    SQLITE_DO_ROLLBACK(db);
    queue_resequence__(db);
    return false;
}

bool
database_queue_add_uri(lkt_db *db, struct lkt_uri *uri, int prio)
{
    const LKT_URI_TYPE type = lkt_uri_get_type(uri);

    switch (type) {
    case URI_QUERY:
        return queue_add_with_col_like_str(db, LKT_DB_ALL, lkt_uri_get_value_as_str(uri), prio);

    case URI_ID: return database_queue_add_id(db, (int)lkt_uri_get_value_as_int(uri), prio);

    case URI_TYPE:
        return queue_add_with_col_like_str(db, LKT_DB_TYPE, lkt_uri_get_value_as_str(uri), prio);

    case URI_CATEGORY:
        return queue_add_with_col_like_str(db, LKT_DB_CAT, lkt_uri_get_value_as_str(uri), prio);

    case URI_LANGUAGE:
        return queue_add_with_col_like_str(db, LKT_DB_LANG, lkt_uri_get_value_as_str(uri), prio);

    case URI_AUTHOR:
        return queue_add_with_col_like_str(db, LKT_DB_AUTHOR, lkt_uri_get_value_as_str(uri), prio);

    case URI_PLAYLIST: return database_queue_add_plt(db, lkt_uri_get_value_as_str(uri), prio);

    case URI_NULL:
        LOG_WARN("DB", "Add to queue for uri of type %d is not implemented", type);
        return false;
    }

    LOG_UNREACHABLE;
}

bool
database_queue_del_id(lkt_db *db, int id)
{
    static const char *SQL_TEMPLATE =
        "BEGIN TRANSACTION;"
        /* Clear the TMP */
        "DELETE FROM queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
        /* Move the current 'pointer' */
        "UPDATE queue_state SET current = (SELECT NULLIF(COUNT(position), 0) FROM queue JOIN queue_state ON position <= current AND kara_id != %d);"
        "DELETE FROM queue WHERE kara_id = %d;" /* Delete any kara with the specified id */
        "INSERT INTO queue_tmp(priority, position, kara_id) SELECT priority, position, kara_id FROM queue ORDER BY position ASC;"
        "DELETE FROM queue;"
        "INSERT INTO queue(priority, position, kara_id) SELECT priority, ROW_NUMBER() OVER(ORDER BY position ASC), kara_id FROM queue_tmp;"
        "UPDATE " LKT_PROTECTED_DATABASE
        ".sqlite_sequence SET seq = (SELECT COUNT(*) FROM queue) WHERE name = 'queue';" /* Update the sqlite_sequence            */
        /* Clear the TMP */
        "DELETE FROM queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
        "COMMIT;";
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    safe_snprintf(SQL, LKT_MAX_SQLITE_STATEMENT, SQL_TEMPLATE, id, id);
    SQLITE_EXEC(db, SQL, error);
    queue_resequence__(db);
    return true;
error:
    SQLITE_DO_ROLLBACK(db);
    return false;
}

bool
database_queue_del_pos(lkt_db *db, int pos)
{
    /* /!\ WARNING: UGLY GOTOS AT THE END OF THE FUNCTION /!\ */

    bool sta                     = false;
    sqlite3_stmt *stmt           = NULL;
    struct lkt_queue_state queue = { .current = -1 };
    RETURN_UNLESS(database_queue_state(db, &queue), "Failed to get the queue state", false);

    /* Delete the current kara, can't do that because we need to skip it, and
     * send signals to the window module to do so, so we just fail here... */
    if (queue.current == pos) {
        LOG_ERROR("DB", "Can't delete the currently playing kara at position %d in queue", pos);
        return false;
    }

    static const char *SQL_TEMPLATE =
        /* Create temporary queue */
        "DELETE FROM queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
        /* -1 to all positions after the deleted kara */
        "DELETE FROM queue WHERE position = %d;"
        "INSERT INTO queue_tmp (kara_id, priority)"
        "  SELECT kara_id, priority"
        "  FROM queue;"
        /* Re-init positions in the queue */
        "DELETE FROM queue;"
        "DELETE FROM " LKT_PROTECTED_DATABASE ".sqlite_sequence WHERE name = 'queue';"
        /* Insert back all karas */
        "INSERT INTO queue (kara_id, priority, position)"
        "  SELECT kara_id, priority, position"
        "  FROM queue_tmp;"
        /* Slap correct values here */
        "UPDATE " LKT_PROTECTED_DATABASE ".sqlite_sequence"
        "  SET seq = (SELECT COUNT(position) FROM queue)"
        "  WHERE name = 'queue';"
        "DELETE FROM queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';";
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    safe_snprintf(SQL, LKT_MAX_SQLITE_STATEMENT, SQL_TEMPLATE, pos);

    SQLITE_EXEC(db, "BEGIN TRANSACTION;", error_no_rollback);
    SQLITE_EXEC(db, SQL, error_rollback);
    if (pos < queue.current) {
        LOG_DEBUG("DB", "Needs to decrement current index: pos %d < current %d", pos,
                  queue.current);
        SQLITE_EXEC(db, "UPDATE queue_state SET current = current - 1;", error_in_delete_or_update);
    }

    sta = true;
error_in_delete_or_update:
    if (sta) {
        /* Commit transaction, if failed use a magnificient goto... */
        queue_resequence__(db);
        SQLITE_EXEC(db, "END TRANSACTION;", error_rollback);
    }

    else {
    error_rollback:
        sta = false;
        LOG_DEBUG("DB", "An error occured, needs to rollback transaction");
        SQLITE_DO_ROLLBACK(db);
    }

error_no_rollback:
    if (stmt != NULL)
        sqlite3_finalize(stmt);
    return sta;
}

/* When a kara has been played, MPD specify that its priority shall be reset,
 * thus we set it to 1 before passing to the next kara. It's not done when
 * passing to the previous kara because the reorder operation would break
 * everything in the queue. */
PRIVATE_FUNCTION bool
___database_queue_reset_current_priority(lkt_db *db)
{
    static const char *SQL_STMT = "UPDATE queue"
                                  "  SET priority = 1"
                                  "  WHERE position = (SELECT current FROM queue_state LIMIT 1);";
    SQLITE_EXEC(db, SQL_STMT, error);
    LOG_DEBUG("DB", "Set priority to 1 for played kara!");
    return true;
error:
    LOG_ERROR("DB", "Failed to reset the priority of played kara!");
    return false;
}

bool
database_queue_next(lkt_db *db, char filepath[PATH_MAX])
{
    static const char *SQL_STMT =
        "SELECT file_path, position, RANDOM()"
        "  FROM kara"
        "  JOIN queue ON kara.id = queue.kara_id"
        "  JOIN queue_state"
        "    ON current IS NULL"
        "    OR (CASE WHEN (SELECT single FROM queue_state LIMIT 1) = 1 THEN position = current"
        "             WHEN (SELECT random FROM queue_state LIMIT 1) = 1 THEN position <> current"
        "             ELSE position > current END)"
        "  ORDER BY CASE WHEN (SELECT random FROM queue_state LIMIT 1) = 1 THEN RANDOM() ELSE 2 END"
        "  LIMIT 1";
    char SQL_UPDATE[LKT_MAX_SQLITE_STATEMENT];
    int code           = SQLITE_OK, id;
    sqlite3_stmt *stmt = NULL;

    SQLITE_BEGIN_TRANSATION(db, error_no_stmt);
    unless (___database_queue_reset_current_priority(db))
        goto error_no_stmt;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);

    code = sqlite3_step(stmt);

    if (code == SQLITE_ROW) {
        id = MAX(1, sqlite3_column_int(stmt, 1));
        safe_snprintf(SQL_UPDATE, LKT_MAX_SQLITE_STATEMENT, "UPDATE queue_state SET current = %d;",
                      id);

        if (filepath != NULL)
            strncpy(filepath, sqlite3_column_chars(stmt, 0), PATH_MAX - 1);
        else {
            LOG_ERROR("DB", "Failed to get file, id was %d", id);
            goto error;
        }
    }

    else if (code == SQLITE_DONE) {
        LOG_ERROR("DB", "Failed to get next");
        goto error;
    }

    else {
        LOG_ERROR("DB", "Failed to fetch next kara: %s", sqlite3_errmsg((sqlite3 *)db));
        goto error;
    }

    SQLITE_EXEC(db, SQL_UPDATE, error);
    SQLITE_END_TRANSATION(db, error);
    sqlite3_finalize(stmt);
    return true;

error:
    sqlite3_finalize(stmt);
    SQLITE_DO_ROLLBACK(db);
    LOG_ERROR("DB", "Failed in the next code");
    return false;

error_no_stmt:
    LOG_ERROR("DB", "Failed because of reset current priority function");
    SQLITE_DO_ROLLBACK(db);
    return false;
}

bool
database_queue_skip_current(lkt_db *db, char filepath[PATH_MAX])
{
    if (database_queue_next(db, filepath))
        return true;
    else if (database_queue_prev(db, filepath))
        return true;
    else {
        database_queue_stop(db);
        return false;
    }
}

/* When playing the previous kara, if the current was a priority 5 and the
 * prvious a priority 1 we will have a problem on the resequence of the queue.
 * Thus, we propagate all the priorities > 1 if found after the current to all
 * the karas from that found kara with a > 1 priority to the current one
 * included. It only makes sense to call that function from the
 * 'database_queue_prev' function. */
PRIVATE_FUNCTION bool
___database_back_propagate_priority(lkt_db *db)
{
    /* Get the next kara with priority > 1 after the current in the queue */
    static const char *SQL_GET_POS_WITH_PRIO_SUP_1 =
        "SELECT queue.position, queue.priority"
        "  FROM queue"
        "  JOIN queue_state ON"
        "    queue_state.current <= queue.position AND queue.priority > 1"
        "    ORDER BY position ASC LIMIT 1;";

    /* Get the currently playing kara */
    static const char *SQL_GET_CURRENT = "SELECT current FROM queue_state LIMIT 1;";

    /* Apply priority to range */
    static const char *SQL_APPLY_PRIORITY_TO_RANGE =
        "UPDATE queue SET priority = ? WHERE position >= ? AND position <= ?;";

    int64_t currently_playing_position = 0;
    int64_t next_kara_with_prio_sup_1  = 0;
    int priority_to_back_propagate     = 1;
    sqlite3_stmt *stmt                 = NULL;

    /* Currently playing position */
    SQLITE_PREPARE(db, stmt, SQL_GET_CURRENT, error);
    SQLITE_STEP_ROW(db, stmt, not_playing_or_no_need_to_back_propagate);
    currently_playing_position = sqlite3_column_int64(stmt, 0);
    sqlite3_finalize(stmt);

    /* Next kara with priority > 1 */
    SQLITE_PREPARE(db, stmt, SQL_GET_POS_WITH_PRIO_SUP_1, error);
    SQLITE_STEP_ROW(db, stmt, not_playing_or_no_need_to_back_propagate);
    next_kara_with_prio_sup_1  = sqlite3_column_int64(stmt, 0);
    priority_to_back_propagate = sqlite3_column_int(stmt, 1);
    sqlite3_finalize(stmt);

    LOG_DEBUG("DB", "Update priorities to %d from position %ld to %ld", priority_to_back_propagate,
              currently_playing_position, next_kara_with_prio_sup_1);

    /* Back-propagate the priority */
    SQLITE_PREPARE(db, stmt, SQL_APPLY_PRIORITY_TO_RANGE, error);
    SQLITE_BIND_INT(db, stmt, 1, priority_to_back_propagate, error);
    SQLITE_BIND_INT64(db, stmt, 2, currently_playing_position, error);
    SQLITE_BIND_INT64(db, stmt, 3, next_kara_with_prio_sup_1, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sqlite3_finalize(stmt);

    /* Everything whent OK */
    return true;

    /* An error occured, free the sqlite statement */
error:
    if (stmt != NULL)
        sqlite3_finalize(stmt);
    return false;

    /* Not a failure and no operations needed:
     * - If not playing no need to back propagate.
     * - No kara with priority > 1 after the playing one, no need to
     *   backpropagate the priority. */
not_playing_or_no_need_to_back_propagate:
    LOG_DEBUG("DB", "Don't need to back propagate priorities!");
    if (stmt != NULL)
        sqlite3_finalize(stmt);
    return true;
}

bool
database_queue_prev(lkt_db *db, char filepath[PATH_MAX])
{
    static const char *SQL_STMT =
        "SELECT file_path, position "
        "  FROM kara"
        "  JOIN queue ON kara.id = queue.kara_id AND available = 1"
        "  JOIN queue_state ON CASE "
        "    WHEN (SELECT single FROM queue_state LIMIT 1) = 1 THEN position = current"
        "    ELSE queue.position < queue_state.current"
        "  END"
        "  ORDER BY position DESC LIMIT 1;";
    char SQL_UPDATE[LKT_MAX_SQLITE_STATEMENT];
    int code           = SQLITE_OK;
    sqlite3_stmt *stmt = NULL;

    SQLITE_BEGIN_TRANSATION(db, error_failed_to_begin_transaction);

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    code = sqlite3_step(stmt);

    if (code == SQLITE_ROW) {
        int id                             = MAX(1, sqlite3_column_int(stmt, 1));
        static const char *SQL_TMPL_UPDATE = "UPDATE queue_state SET current = %d;";
        safe_snprintf(SQL_UPDATE, LKT_MAX_SQLITE_STATEMENT, SQL_TMPL_UPDATE, id);

        if (filepath != NULL)
            strncpy(filepath, sqlite3_column_chars(stmt, 0), PATH_MAX - 1);
        else {
            LOG_ERROR("DB", "Failed to get file, position was %d", id);
            goto error;
        }
    }

    else if (code == SQLITE_DONE) {
        LOG_ERROR("DB", "Failed to get previous");
        goto error;
    }

    else {
        LOG_ERROR("DB", "Failed to fetch prev kara: %s", sqlite3_errmsg((sqlite3 *)db));
        goto error;
    }

    SQLITE_EXEC(db, SQL_UPDATE, error);
    unless (___database_back_propagate_priority(db))
        goto error;
    sqlite3_finalize(stmt);
    SQLITE_END_TRANSATION(db, error);
    return true;

error:
    SQLITE_DO_ROLLBACK(db);
    sqlite3_finalize(stmt);
    return false;

error_failed_to_begin_transaction:
    return false;
}

bool
database_queue_swap(lkt_db *db, int from, int to)
{
    static const char *SQL_POP_ROW_1 =
        "SELECT position, priority, kara_id FROM queue WHERE position = ?;";
    static const char *SQL_POP_ROW_2 = "DELETE FROM queue WHERE position = ?;";
    static const char *SQL_SLAP_MAX  = "UPDATE " LKT_PROTECTED_DATABASE ".sqlite_sequence"
                                       "  SET seq = (SELECT COUNT(position) FROM queue)"
                                       "  WHERE name = 'queue';";
    static const char *SQL_PUSH_ROW =
        "INSERT INTO queue (position, priority, kara_id) VALUES (?, ?, ?);";

    struct {
        int position;
        int priority;
        int kara_id;
    } kara_from, kara_to;
    sqlite3_stmt *stmt = NULL;
    bool sta           = false;
    int tmp_kara_id;

    /* Get one kara from the 'queue' and delete it */
#define POP_QUEUE(name)                                 \
    SQLITE_PREPARE(db, stmt, SQL_POP_ROW_1, error);     \
    SQLITE_BIND_INT(db, stmt, 1, name, error);          \
    SQLITE_STEP_ROW(db, stmt, error);                   \
    kara_##name.position = sqlite3_column_int(stmt, 0); \
    kara_##name.priority = sqlite3_column_int(stmt, 1); \
    kara_##name.kara_id  = sqlite3_column_int(stmt, 2); \
    sqlite3_finalize(stmt);                             \
    SQLITE_PREPARE(db, stmt, SQL_POP_ROW_2, error);     \
    SQLITE_BIND_INT(db, stmt, 1, name, error);          \
    SQLITE_STEP_DONE(db, stmt, error);                  \
    sqlite3_finalize(stmt);
    /* Push one kara on the 'queue' */
#define PUSH_QUEUE(name)                                       \
    SQLITE_PREPARE(db, stmt, SQL_PUSH_ROW, error);             \
    SQLITE_BIND_INT(db, stmt, 1, kara_##name.position, error); \
    SQLITE_BIND_INT(db, stmt, 2, kara_##name.priority, error); \
    SQLITE_BIND_INT(db, stmt, 3, kara_##name.kara_id, error);  \
    SQLITE_STEP_DONE(db, stmt, error);                         \
    sqlite3_finalize(stmt);

    SQLITE_EXEC(db, "BEGIN TRANSACTION;", error_no_rollback);
    POP_QUEUE(from);
    POP_QUEUE(to);
    tmp_kara_id       = kara_from.kara_id;
    kara_from.kara_id = kara_to.kara_id;
    kara_to.kara_id   = tmp_kara_id;
    PUSH_QUEUE(from);
    PUSH_QUEUE(to);
    SQLITE_EXEC(db, SQL_SLAP_MAX, error_no_stmt);
    SQLITE_EXEC(db, "END TRANSACTION;", error_no_stmt);

#undef PUSH_QUEUE
#undef POP_QUEUE

    sta = true;
    goto error_no_stmt;
error:
    sqlite3_finalize(stmt);
error_no_stmt:
    if (!sta)
        SQLITE_DO_ROLLBACK(db);
error_no_rollback:
    return sta;
}

bool
database_queue_move(lkt_db *db, int from, int to)
{
    static const char *SQL_TEMPLATE_PART_1 =
        /* Clean queue_tmp */
        "DELETE FROM queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
        /* Construct queue */
        "INSERT INTO queue_tmp(kara_id,priority)"
        "WITH TEMP_TABLE AS (SELECT"
        "  position, kara_id, priority"
        "  FROM queue"
        "  ORDER BY position ASC, priority DESC)"
        "SELECT kara_id, priority"
        "  FROM TEMP_TABLE WHERE POSITION < %d AND POSITION != %d ORDER BY POSITION;"
        /* Construct queue */
        "INSERT INTO queue_tmp(kara_id,priority)"
        "  VALUES(("
        "    WITH TEMP_TABLE AS ("
        "      SELECT ROW_NUMBER() OVER(ORDER BY position ASC, priority DESC) AS POSITION,"
        "        kara_id, priority FROM queue)"
        "      SELECT kara_id FROM TEMP_TABLE WHERE POSITION = %d"
        "    ),("
        "      WITH TEMP_TABLE AS ("
        "      SELECT ROW_NUMBER() OVER(ORDER BY position ASC, priority DESC) AS POSITION,"
        "        kara_id, priority FROM queue)"
        "      SELECT priority FROM TEMP_TABLE WHERE POSITION = "
        "      CASE WHEN (SELECT COUNT(*) FROM TEMP_TABLE)>=%d THEN %d ELSE %d-1 END"
        "  ));";
    static const char *SQL_TEMPLATE_PART_2 =
        /* Construct queue */
        "INSERT INTO queue_tmp(kara_id,priority)"
        "WITH TEMP_TABLE AS ("
        "  SELECT ROW_NUMBER() OVER(ORDER BY position ASC, priority DESC) AS POSITION,"
        "    kara_id, priority FROM queue)"
        "SELECT kara_id, priority"
        "  FROM TEMP_TABLE WHERE POSITION >= %d AND POSITION != %d ORDER BY POSITION;"
        /* Clean queue and insert back */
        "UPDATE queue_state"
        "  SET current = CASE"
        "    WHEN current NOT NULL "
        "    THEN (SELECT position FROM queue_tmp "
        "       WHERE kara_id=(SELECT kara_id FROM queue WHERE position = current))"
        "    ELSE NULL END;"
        "DELETE FROM queue;"
        "DELETE FROM " LKT_PROTECTED_DATABASE ".sqlite_sequence WHERE name = 'queue';"
        "INSERT INTO queue(position, kara_id, priority)"
        "  SELECT position, kara_id, priority FROM queue_tmp ORDER BY priority, position;"
        /* Clean queue_tmp */
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"

        "DELETE FROM queue_tmp;";

    char SQL[LKT_MAX_SQLITE_STATEMENT];

    SQLITE_EXEC(db, "BEGIN TRANSACTION;", error_no_rollback);
    safe_snprintf(SQL, LKT_MAX_SQLITE_STATEMENT, SQL_TEMPLATE_PART_1, to, from, from, to, to, to);
    SQLITE_EXEC(db, SQL, error);

    safe_snprintf(SQL, LKT_MAX_SQLITE_STATEMENT, SQL_TEMPLATE_PART_2, to, from);
    SQLITE_EXEC(db, SQL, error);
    SQLITE_EXEC(db, "COMMIT;", error);
    return true;
error:
    SQLITE_DO_ROLLBACK(db);
error_no_rollback:
    return false;
}

bool
database_queue_play(lkt_db *db, int pos)
{
    static const char *SQL_STMT = "UPDATE queue_state SET current = ?, paused = 0"
                                  "  WHERE current <= 0 OR current IS NULL";
    int code                    = SQLITE_OK;
    bool ret                    = false;
    sqlite3_stmt *stmt          = NULL;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, pos, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK && code != SQLITE_DONE) {
        LOG_ERROR("DB", "Failed to update queue_state: %s", sqlite3_errmsg((sqlite3 *)db));
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_queue_set_current_index(lkt_db *db, int idx)
{
    static const char *SQL_GET_TEMPLATE = "UPDATE queue_state SET current = %d;";
    char SQL_GET[LKT_MAX_SQLITE_STATEMENT];

    if (idx <= 0) {
        LOG_ERROR("DB", "An idx of %d is invalid, must be >= 0", idx);
        return false;
    }

    if (safe_snprintf(SQL_GET, LKT_MAX_SQLITE_STATEMENT, SQL_GET_TEMPLATE, idx) < 0)
        return false;

    SQLITE_EXEC(db, SQL_GET, error);
    return true;

error:
    return false;
}

bool
database_queue_get_current_file(lkt_db *db, char filepath[PATH_MAX])
{
    static const char *SQL_STMT = "SELECT file_path"
                                  "  FROM kara"
                                  "  JOIN queue ON kara.id = queue.kara_id"
                                  "  JOIN queue_state ON queue.position = queue_state.current";
    bool status                 = false;
    int code                    = SQLITE_OK;
    sqlite3_stmt *stmt          = NULL;

    if (filepath == NULL)
        goto error;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    code = sqlite3_step(stmt);

    if (code == SQLITE_ROW) {
        strncpy(filepath, sqlite3_column_chars(stmt, 0), PATH_MAX);
        filepath[PATH_MAX - 1] = '\0';
    }

    else {
        LOG_ERROR("DB", "Failed to fetch prev kara: %s", sqlite3_errmsg((sqlite3 *)db));
        goto error;
    }

    status = true;

error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_get_kara_path_no_cache(lkt_db *db, int id, char filepath[PATH_MAX])
{
    bool ret_code          = false;
    sqlite3_stmt *stmt     = NULL;
    static const char *SQL = "SELECT file_path FROM kara WHERE id = ?;";
    RETURN_UNLESS(id && db, "Invalid argument", false);
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (filepath != NULL)
        safe_strncpy(filepath, sqlite3_column_chars(stmt, 0), PATH_MAX);
    ret_code = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}

bool
database_get_kara_path(lkt_db *db, int id, char filepath[PATH_MAX])
{
    bool ret_code          = false;
    sqlite3_stmt *stmt     = NULL;
    static const char *SQL = "SELECT file_path FROM kara_cache WHERE kara_id = ?;";
    RETURN_UNLESS(id && db, "Invalid argument", false);
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (filepath != NULL)
        safe_strncpy(filepath, sqlite3_column_chars(stmt, 0), PATH_MAX);
    ret_code = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}

bool
database_get_kara_id(lkt_db *db, char filepath[PATH_MAX], int *id)
{
    bool ret_code          = false;
    sqlite3_stmt *stmt     = NULL;
    static const char *SQL = "SELECT id FROM kara WHERE file_path = ? AND available = 1;";
    RETURN_UNLESS(id && db, "Invalid argument", false);
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_TEXT(db, stmt, 1, filepath, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (id)
        *id = sqlite3_column_int(stmt, 0);
    ret_code = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}

bool
database_get_kara_position(lkt_db *db, int id, int *pos)
{
    bool ret_code          = false;
    sqlite3_stmt *stmt     = NULL;
    static const char *SQL = "WITH content AS ("
                             " SELECT kara.id AS id, position"
                             "  FROM queue"
                             "  JOIN kara ON kara_id = kara.id"
                             "  GROUP BY position ORDER BY position ASC, priority DESC)"
                             "SELECT id, position FROM content WHERE id = ?;";
    RETURN_UNLESS(db && pos, "Invalid argument", false);
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *pos     = sqlite3_column_int(stmt, 0);
    ret_code = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}

int
database_queue_probe_id(lkt_db *db, int id)
{
    int position           = 0; /* The queue begins with the position 1 */
    sqlite3_stmt *stmt     = NULL;
    static const char *SQL = "SELECT position FROM queue WHERE kara_id = ?;";
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    position = sqlite3_column_int(stmt, 0);
error:
    sqlite3_finalize(stmt);
    return position;
}

bool
database_queue_shuffle(lkt_db *db)
{
    const char *SQL = "BEGIN TRANSACTION;"
                      /* Create temporary queue */
                      "DELETE FROM queue_tmp;"
                      "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
                      /* When current is NULL, that thing is also NULL, so no insertion is done */
                      "INSERT INTO queue_tmp (kara_id, priority)"
                      "  SELECT kara_id, 5"
                      "  FROM queue"
                      "  JOIN queue_state ON queue.position = queue_state.current;"
                      /* Insert the others karas */
                      "INSERT INTO queue_tmp (kara_id, priority)"
                      "  SELECT kara_id, priority"
                      "  FROM queue"
                      "  JOIN queue_state ON"
                      "    queue.position <> CASE"
                      "      WHEN queue_state.current IS NOT NULL THEN queue_state.current"
                      "      ELSE 0 END"
                      "  ORDER BY RANDOM();"
                      /* Do the insertion in the other way */
                      "DELETE FROM queue;"
                      "DELETE FROM " LKT_PROTECTED_DATABASE ".sqlite_sequence WHERE name = 'queue';"
                      "INSERT INTO queue (position, kara_id, priority)"
                      "  SELECT position, kara_id, priority"
                      "  FROM queue_tmp"
                      "  ORDER BY priority, position;"
                      /* Set the current to the right kara when needed */
                      "UPDATE queue_state"
                      "  SET current = CASE"
                      "    WHEN current NOT NULL THEN 1"
                      "    ELSE NULL END;"
                      /* Clear the TMP */
                      "DELETE FROM queue_tmp;"
                      "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
                      "COMMIT;";
    database_queue_flat(db);
    SQLITE_EXEC(db, SQL, error);
    return true;
error:
    SQLITE_DO_ROLLBACK(db);
    return false;
}

bool
database_queue_list(lkt_db *db, size_t from, size_t to, struct lkt_callback *callback)
{
    const char *SQL_STMT =
        "WITH content AS ("
        " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len, position, LENGTH(CAST(position AS TEXT)) AS pos_len"
        "  FROM queue"
        "  JOIN kara ON kara_id = kara.id"
        "  WHERE position >= ? AND position <= ?"
        "  GROUP BY position ORDER BY position ASC, priority DESC)"
        "SELECT id, string, (SELECT MAX(len) FROM content), position, (SELECT MAX(pos_len) FROM content) FROM content;";
    int code, id, id_len, pos, pos_len;
    const char *row;
    bool ret = false;
    sqlite3_stmt *stmt;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, (int)from, error);
    SQLITE_BIND_INT(db, stmt, 2, (int)to, error);

    for (;;) {
        code = sqlite3_step(stmt);

        if (code == SQLITE_ROW) {
            id      = sqlite3_column_int(stmt, 0);
            row     = sqlite3_column_chars(stmt, 1);
            id_len  = sqlite3_column_int(stmt, 2);
            pos     = sqlite3_column_int(stmt, 3);
            pos_len = sqlite3_column_int(stmt, 4);
            ++(callback->iterations);
            if (callback->call(callback, pos_len, pos, id, id_len, row))
                continue;
            else
                break;
        }

        else if (code == SQLITE_OK || code == SQLITE_DONE)
            goto done;

        else {
            LOG_ERROR("DB", "Failed: %s", sqlite3_errmsg((sqlite3 *)db));
            break;
        }
    }

done:
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_queue_dump(lkt_db *db, const char *plt_name)
{
    static const char *SQL = "WITH plt_id AS (SELECT playlist.id AS id FROM playlist"
                             " WHERE name COLLATE nocase = ?) "
                             "INSERT OR IGNORE INTO kara_playlist (kara_id, playlist_id)"
                             " SELECT DISTINCT kara_id, plt_id.id"
                             " FROM queue, plt_id"
                             " ORDER BY RANDOM();";
    bool ret_code          = false;
    sqlite3_stmt *stmt     = NULL;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_TEXT(db, stmt, 1, plt_name, error);
    SQLITE_STEP_DONE(db, stmt, error);
    ret_code = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}

bool
database_queue_seekid(lkt_db *db, int id, int *out_pos)
{
    static const char *SQL_STMT = "SELECT position FROM queue WHERE kara_id = ? LIMIT 1";
    int ret                     = 0;
    sqlite3_stmt *stmt;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *out_pos = sqlite3_column_int(stmt, 0);
    ret      = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_queue_multi_add_id(lkt_db *db, size_t *ids, int priority)
{
    if (ids == NULL) {
        LOG_ERROR("DB", "Can't add a list of IDs to the queue when the list's pointer is NULL");
        return false;
    }
    if ((priority < 1) || (priority > 5)) {
        LOG_ERROR("DB", "The priority must be between 1 and 5, here it is %d", priority);
        return false;
    }

    static const size_t ADD_STEP = 10; /* TODO: Let the user specify that thing */
    static const char *SQL_STMT_TEMPLATE =
        "INSERT INTO queue (kara_id, priority)"
        " SELECT kara.id, %d"              /* The priority */
        " FROM kara WHERE kara.id IN (%s)" /* The list of ids, comma-separated */
        ";";

    const size_t ids_len = arrlenu(ids);
    size_t id_index      = 0;
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT]; /* The formated query to add things to the queue */
    char IDS_LIST[LKT_LINE_MAX]; /* The thing that will hold the formated comma-separated list of ids */

    SQLITE_BEGIN_TRANSATION(db, error);

    while (id_index < ids_len) {
        const size_t remaining    = MIN(ids_len - id_index, ADD_STEP);
        const size_t cat_loop_lim = id_index + remaining;
        size_t index_in_id_string = 0;

        memset(IDS_LIST, 0, sizeof(IDS_LIST));
        memset(SQL_STMT, 0, sizeof(SQL_STMT));

        if (remaining == 0) {
            LOG_WARN("DB", "Don't add any more IDs because none are remaining. "
                           "Be aware that this part of the code should not be executed but "
                           "only a warning is raised as it has no incidence on the state of "
                           "the program");
            break;
        }

        for (/* Nothing */; id_index < cat_loop_lim; ++id_index) {
            char *const addr      = IDS_LIST + index_in_id_string;
            const size_t max_size = LKT_LINE_MAX - 1 - index_in_id_string;
            const int added       = safe_snprintf(addr, max_size, "%zu,", ids[id_index]);
            if (added <= 0) {
                /* Ouups... */
                LOG_FATAL("Failed to concatenate something to the list string, "
                          "partial state is: %s",
                          IDS_LIST);
            }
            index_in_id_string += (size_t)added;
        }

        /* Delete the last comma */
        IDS_LIST[index_in_id_string - 1] = '\0';
        LOG_DEBUG("DB", "Add the following IDs to queue with priority %d: %s", priority, IDS_LIST);
        safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TEMPLATE, priority, IDS_LIST);
        SQLITE_EXEC(db, SQL_STMT, error);
    }

    reorder(db, priority, error);
    SQLITE_END_TRANSATION(db, error);
    return true;

error:
    SQLITE_DO_ROLLBACK(db);
    return false;
}

/* Initialize the queue. Can't fail if the passed argument is not NULL.
 * WARN: In case of out of memory this function will abort the process. */
int
lkt_queue_new(struct queue *ret)
{
    if (!ret)
        return 1;

    pthread_mutex_t mxt = PTHREAD_MUTEX_INITIALIZER;
    struct queue _ret   = {
          .contents  = LKT_ALLOC_ARRAY(lkt_event, LKT_DEFAULT_LIST_SIZE),
          .size      = LKT_DEFAULT_LIST_SIZE,
          .last      = 0,
          .available = 0,
          .lock      = mxt,
    };

    *ret = _ret;
    return 0;
}

/* Free the queue. Can't fail. */
void
lkt_queue_free(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    if (queue && queue->contents)
        free((void *)queue->contents);
    pthread_mutex_unlock(&queue->lock);
}

/* Is an event of the following type present in the queue?
 * Returns 0 if the type is not present, 1 otherwise. */
bool
lkt_queue_has_event(struct queue *queue, LKT_EVENT_TYPE type)
{
    bool ret = false;
    pthread_mutex_lock(&queue->lock);
    for (size_t i = 0; i < queue->last; ++i) {
        if (queue->contents[i].type & type) {
            ret = true;
            break;
        }
    }
    pthread_mutex_unlock(&queue->lock);
    return ret;
}

/* Send an event into the queue */
void
lkt_queue_send(struct queue *queue, LKT_EVENT_TYPE _type, void *_attr)
{
    pthread_mutex_lock(&queue->lock);
    GOTO_UNLESS(queue, "Invalid argument", end)

    volatile lkt_event *new;
    if (queue->size == queue->last) {
        new = realloc((void *)queue->contents, queue->size * 2 * sizeof(lkt_event));

        if (NULL == new)
            goto end;

        queue->contents = new;
        queue->size *= 2;
    }

    lkt_event evt = {
        .type = _type,
        .attr = _attr,
    };
    queue->contents[(queue->last)++] = evt;

    if (!(_type & (unsigned int)queue->available))
        LOG_WARN("QUEUE", "The event %d is not available, push it anyway", _type);
end:
    pthread_mutex_unlock(&queue->lock);
}

lkt_event
lkt_queue_handle(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    lkt_event ret = { 0 };
    if (!queue || !queue->last)
        goto end;
    if (!(queue->contents[0].type & (unsigned int)queue->available) && queue->contents[0].type) {
        LOG_WARN("QUEUE", "Event %d is not available", queue->contents[0].type);
        goto end;
    }

    ret = queue->contents[0];
    for (size_t i = 1; i < queue->last; ++i)
        queue->contents[i - 1] = queue->contents[i];
    queue->last -= 1;
end:
    pthread_mutex_unlock(&queue->lock);
    return ret;
}

void
lkt_queue_make_available(struct queue *queue, LKT_EVENT_TYPE type)
{
    pthread_mutex_lock(&queue->lock);
    queue->available |= type;
    pthread_mutex_unlock(&queue->lock);
}
