import ycm_core
from os import getcwd
from os.path import abspath, join, isabs, normpath, exists, splitext, dirname

default_flags           = [ "-Wall", "-Wextra" ]
c_header_extensions     = [ ".h" ]
c_source_extensions     = [ ".c" ]
c_additional_flags      = [ "-x", "c", "-std=c11", "-I", "./inc/" ]
cpp_header_extensions   = [ ".h", ".hh", ".H", ".hp", ".hpp", ".HPP", ".hxx", ".h++" ]
cpp_source_extensions   = [ ".cp", ".cpp", ".CPP", ".cc", ".C", ".cxx", ".c++" ]
cpp_additional_flags    = [ "-x", "c++", "-std=c++17" ]

def find_file_recursively(file_name, start_dir = getcwd(), stop_dir = None):
    cur_dir = abspath(start_dir) if not isabs(start_dir) else start_dir
    while True:
        if exists(join(cur_dir, file_name)):
            return join(cur_dir, file_name)

        parent_dir = dirname(cur_dir)

        if parent_dir == cur_dir or parent_dir == stop_dir:
            return None

        else:
            cur_dir = parent_dir


def file_exists(file_name, start_dir = getcwd()):
    return find_file_recursively(file_name, start_dir) is not None

def make_path_absolute(path, base_dir=getcwd()):
    if isabs(path): return path
    else:           return join(base_dir, path)

def script_directory():
    return dirname(__file__)

def is_header(file_path):
    return is_c_header(file_path) or is_cpp_header(file_path)

def is_c_header(file_path):
    (_, extension) = splitext(file_path)
    return extension in c_header_extensions

def is_cpp_header(file_path):
    (_, extension) = splitext(file_path)
    return extension in cpp_header_extensions

def is_source(file_path):
    return is_c_source(file_path) or is_cpp_source(file_path)

def is_c_source(file_path):
    (_, extension) = splitext(file_path)
    return extension in c_source_extensions

def is_cpp_source(file_path):
    (_, extension) = splitext(file_path)
    return extension in cpp_source_extensions

def is_c_file(file_path):
    return is_c_source(file_path) or is_c_header(file_path)

def is_cpp_file(file_path):
    return is_cpp_source(file_path) or is_cpp_header(file_path)

def make_absolute_flags(flags, base_dir):
    next_is_path     = [ "-I", "-isystem", "-iquote" ]
    argument_is_path = [ "--sysroot=" ]
    updated_flags    = []
    make_absolute    = False

    for flag in flags:
        updated_flag = flag

        if make_absolute:
            updated_flag = make_path_absolute(flag, base_dir)
            make_absolute = False

        if flag in next_is_path:
            make_absolute = True

        for f in next_is_path:
            if flag.startswith(f):
                path = flag[len(f):].lstrip()
                updated_flags.append(f)
                updated_flag = make_path_absolute(path, base_dir)
                break

        for f in argument_is_path:
            if flag.startswith(f):
                path = flag[len(f):].lstrip()
                updated_flag = f + make_path_absolute(path, base_dir)
                break

        updated_flags.append(updated_flag)

    return updated_flags

def strip_flags(flags):
    return [flag.strip() for flag in flags]

def make_final_flags(file_name, flags, base_dir = getcwd()):
    stripped = strip_flags(flags)
    absolute = make_absolute_flags(stripped, base_dir)

    if is_cpp_file(file_name):  final = save_add_flags(absolute, cpp_additional_flags)
    elif is_c_file(file_name):  final = save_add_flags(absolute, c_additional_flags)
    else:                       final = absolute

    return create_result(final)

def save_add_flags(old_flags, additional_flags):
    skip_next_af = False

    for j in range(len(additional_flags)):
        af            = additional_flags[j].strip()
        argument_type = "none"
        to_add        = True

        if skip_next_af:
            skip_next_af = False
            continue

        if len(additional_flags) > j + 1 and not additional_flags[j+1].startswith("-"):
            argument_type   = "next"
            skip_next_af    = True
            af_arg          = additional_flags[j+1].strip()

        elif af.find(" ") != -1:
            pos             = af.find(" ")
            argument_type   = "next"
            af_arg          = af[pos+1:].strip()
            af              = af[:pos]

        elif af.find("=") != -1:
            pos             = af.find("=")
            argument_type   = "same"
            af_arg          = af[pos+1:].strip()
            af              = af[:pos]

        skip_next_of = False

        for i in range(len(old_flags)):
            of = old_flags[i].strip()

            if skip_next_of:
                skip_next_of = False
                continue

            if argument_type == "none":
                if of == af:
                    to_add = False
                    break

            elif argument_type == "next":
                if of == af:
                    to_add = False
                    skip_next_of = True
                    break

                elif of.startswith(af) and of[len(af)] == " ":
                    to_add = False
                    break

                elif of.startswith(af):
                    to_add = False
                    break

            elif argument_type == "same":
                if of.startswith(af):
                    if len(of) > len(af) + 1 and of[len(af)] == "=":
                        to_add = False
                        break

                    elif len(of) == len(af) + 1 and of[len(af)] == "=":
                        to_add = False
                        skip_next_of = True
                        break

                    elif len(of) == len(af) and len(old_flags) > i + 1 and old_flags[i+1].strip().startswith("="):
                        to_add = False
                        skip_next_of = True
                        break

        if to_add:
            if   argument_type == "none": old_flags.append(af)
            elif argument_type == "next": old_flags.extend([af, af_arg])
            elif argument_type == "same": old_flags.append("{}={}".format(af, af_arg))

    return old_flags

def create_result(flags, do_cache = True, **kwargs):
    ret = {"flags": flags, "do_cache": do_cache}
    return dict(ret, **kwargs)

def parse_compile_commands(file_name, search_base = getcwd()):
    database_path = dirname(find_file_recursively("compile_commands.json", search_base))
    database = ycm_core.CompilationDatabase(database_path)

    if is_header(file_name):
        (name,_) = splitext(file_name)

        for ext in (c_source_extensions + cpp_source_extensions):
            alternative_name = name + ext

            if exists(alternative_name):
                compilation_info = database.GetCompilationInfoForFile(alternative_name)

                if (compilation_info.compiler_flags_):
                    return make_final_flags(file_name, compilation_info.compiler_flags_,
                            compilation_info.compiler_working_dir_)

    elif is_source(file_name):
        compilation_info = database.GetCompilationInfoForFile(file_name)

        if (compilation_info.compiler_flags_):
            return make_final_flags(file_name, compilation_info.compiler_flags_, compilation_info.compiler_working_dir_)

    return parse_default_flags(file_name)

def parse_clang_complete(file_name, search_base = getcwd()):
    config = find_file_recursively(".clang_complete", search_base)
    config_path = dirname(config)

    with open(config, "r") as config_file:
        flags = config_file.read().splitlines()
        return make_final_flags(file_name, flags, config_path)

def parse_default_flags(file_name):
    return make_final_flags(file_name, default_flags, script_directory())

def FlagsForFile(file_name, **kwargs):
    search_base = dirname(file_name)
    if file_exists("compile_commands.json", search_base):   return parse_compile_commands(file_name, search_base)
    elif file_exists(".clang_complete", search_base):       return parse_clang_complete(file_name, search_base)
    else:                                                   return parse_default_flags(file_name)
